/*
 * CFS (Corba File System) for Linux.
 * Livio Soares <livio@ime.usp.br>
 * http://www.lcpd.ime.usp.br/~livio/CFS/
 *
 * This file is released under the GPL.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/devfs_fs_kernel.h>

#include <asm/uaccess.h>

#include <linux/cfs.h>

static struct super_operations cfs_ops;
struct address_space_operations cfs_symlink_aops;
struct inode_operations cfs_symlink_inode_operations;
static struct address_space_operations cfs_aops;
static struct inode_operations cfs_file_inode_operations;
static struct file_operations cfs_file_operations;
static struct inode_operations cfs_dir_inode_operations;
static struct file_operations cfs_dir_operations;
struct dentry_operations cfs_dentry_operations;

static int cfs_statfs(struct super_block *sb, struct statfs *buf)
{
	buf->f_type = CFS_MAGIC;
	buf->f_bsize = PAGE_CACHE_SIZE;
	buf->f_namelen = CFS_NAME_LEN;

	if (ll_cfs_statfs(sb, buf)) {
		/* fake something like AFS & CODA does */
		buf->f_blocks = 9000000;
		buf->f_bfree  = 9000000;
		buf->f_bavail = 9000000;
		buf->f_files  = 9000000;
		buf->f_ffree  = 9000000;
	}
	return 0;
}

/*
 * Lookup the data. This is trivial - if the dentry didn't already
 * exist, we know it is negative.
 */
static struct dentry * cfs_lookup(struct inode *dir, struct dentry *dentry)
{
	int error;
	struct inode *inode = NULL;

	if (dentry->d_inode)
		inode = dentry->d_inode;

	error = ll_cfs_lookup(&inode, dentry);

	/* real error */
	if (error < 0 && error != -ENOENT) {
		printk(".. error!\n");
		d_drop(dentry);
		return ERR_PTR(error);
	}

	if (error == -ENOENT) {
		printk(".. not found (ENOENT)!\n");
		/*dentry->d_time = 0;*/
                dentry->d_op = &cfs_dentry_operations;
		/*if (inode)
		  d_add(dentry, NULL); */

		/*return ERR_PTR(-ENOENT);*/
		return ERR_PTR(0);
	}

	if (inode) {
		printk(".. found it!\n");
		/*dentry->d_time = 0;*/
		dentry->d_op = &cfs_dentry_operations;
		d_add(dentry, inode);
		dget(dentry);
		/*dput(dentry);		   Extra count - pin the dentry in core */
		/* iput(inode); */
	}
	else {
		printk(".. not found!\n");
		d_drop(dentry);
		
		/*velho*/
		/* d_add(dentry, NULL); */
		/* dput(dentry); */
/*		return (ERR_PTR(-ENOENT));
		return dentry; */
	}
	return ERR_PTR(0);
}

int cfs_lookup_revalidate(struct dentry * dentry, int flags)
{
	/* new or old dentry? */

	if (dentry->d_inode) {
		int error;
		/* clean the pages */
		/*truncate_inode_pages(dentry->d_inode->i_mapping, 0);*/ 
		error = ll_cfs_lookup(&(dentry->d_inode), dentry);
		
		printk("CFS: cfs_lookup_revalidate: %s:%s\n", dentry->d_name.name,
		       (!error ? "TRUE" : "FALSE"));
		if (!error) {
			printk(".. found it!\n");
			return 1;
		}
		printk(".. not found!\n");
		shrink_dcache_parent(dentry);
		/*dput(dentry);*/
		d_drop(dentry);
		return 0;
	}
	else {
		struct dentry* error = cfs_lookup(dentry->d_inode, dentry);
		printk("CFS: cfs_lookup_revalidate: %s:%s\n", dentry->d_name.name, 
		       (!error ? "TRUE" : "FALSE"));
/*		if (error)
		return 1;
		return 0; */
		
		return PTR_ERR(error);
	}
}

/*
 * Read a page. Again trivial. If it didn't already exist
 * in the page cache, it is zero-filled.
 */
static int cfs_readpage(struct file *file, struct page * page)
{
	int error;
	char *buffer;

	get_page(page);
	LockPage(page);
	ClearPageUptodate(page);
	ClearPageError(page);

	buffer = kmap(page);

	/*memset(buffer, 0, PAGE_CACHE_SIZE);*/
	
	error = ll_cfs_readpage(file->f_dentry, ((loff_t)page->index << PAGE_CACHE_SHIFT), buffer);

	flush_dcache_page(page);
	SetPageUptodate(page);
	kunmap(page);
	UnlockPage(page);
	

	return error;
}

static int cfs_prepare_write(struct file *file, struct page *page, unsigned offset, unsigned to)
{
        void *addr = kmap(page);
        if (!Page_Uptodate(page)) {
                memset(addr, 0, PAGE_CACHE_SIZE);
                flush_dcache_page(page);
                SetPageUptodate(page);
        } 
        SetPageDirty(page);
        return 0;
}

static int cfs_commit_write(struct file *file, struct page *page, unsigned offset, unsigned to)
{
        struct inode *inode = page->mapping->host;
        loff_t pos = ((loff_t)page->index << PAGE_CACHE_SHIFT) + to;
	int error;

	if ((error = ll_cfs_commit_write(file, page, offset, to))) {
		kunmap(page);
		return error;
	}
	
        kunmap(page);
        if (pos > inode->i_size)
                inode->i_size = pos;
	return 0;
}

void cfs_get_inode_define_mode(struct inode *inode)
{
	switch (inode->i_mode & S_IFMT) {
	default:
		init_special_inode(inode, inode->i_mode, inode->i_dev);
		break;
	case S_IFREG:
		inode->i_fop = &cfs_file_operations;
		inode->i_op = &cfs_file_inode_operations;
		break;
	case S_IFDIR:
		inode->i_op = &cfs_dir_inode_operations;
		inode->i_fop = &cfs_dir_operations;
		break;
	case S_IFLNK:
		inode->i_op = &cfs_symlink_inode_operations;
		inode->i_data.a_ops = &cfs_symlink_aops;
                inode->i_mapping = &inode->i_data;
		break;
	}
}


struct inode *cfs_get_inode(struct super_block *sb, int mode, int dev)
{
	struct inode * inode = new_inode(sb);

	if (inode) {
		inode->i_mode = mode;
		inode->i_uid = current->fsuid;
		inode->i_gid = current->fsgid;
		inode->i_blksize = PAGE_CACHE_SIZE;
		inode->i_blocks = 0;
		inode->i_dev = dev;
		inode->i_rdev = NODEV;
		inode->i_mapping->a_ops = &cfs_aops;
		inode->i_atime = inode->i_mtime = inode->i_ctime = CURRENT_TIME;
		if (mode)
			cfs_get_inode_define_mode(inode);	
	}
	return inode;
}

void cfs_force_delete(struct inode *inode)
{
	/*
	 * Kill off unused inodes ... iput() will unhash and
	 * delete the inode if we set i_nlink to zero.
	 */
	if (atomic_read(&inode->i_count) == 1)
		inode->i_nlink = 0;
	ll_cfs_force_delete(inode);
}

/*
 * File creation. Allocate an inode, and we're done..
 */
static int cfs_mknod(struct inode *dir, struct dentry *dentry, int mode, int dev)
{
	int error = -ENOSPC;
	struct dentry *entry;

	if ((error = ll_cfs_mknod(dir, dentry, mode, dev))) {
		return error;
	}
	
	if ((entry  = cfs_lookup(dir, dentry))) {
		return PTR_ERR(dentry);
	}

        return 0;

}

static int cfs_mkdir(struct inode * dir, struct dentry * dentry, int mode)
{
	return cfs_mknod(dir, dentry, mode | S_IFDIR, 0);
}

static int cfs_create(struct inode *dir, struct dentry *dentry, int mode)
{
	return cfs_mknod(dir, dentry, mode | S_IFREG, 0);
}

/*
 * Link a file..
 */
static int cfs_link(struct dentry *old_dentry, struct inode * dir, struct dentry * dentry)
{
	struct inode *inode = old_dentry->d_inode;

	if (S_ISDIR(inode->i_mode))
		return -EPERM;

	inode->i_nlink++;
	atomic_inc(&inode->i_count);	/* New dentry reference */
	dget(dentry);		/* Extra pinning count for the created dentry */
	d_instantiate(dentry, inode);
	ll_cfs_link(old_dentry, dir, dentry);
	return 0;
}

/*
 * This works for both directories and regular files.
 * (non-directories will always have empty subdirs)
 */
static int cfs_unlink(struct inode * dir, struct dentry *dentry)
{
	int retval;

	retval = ll_cfs_unlink(dir, dentry);	

	if (!retval) {
		struct inode *inode = dentry->d_inode;
		
		inode->i_nlink--;
		dput(dentry);
	}
	return retval;
}

#define cfs_rmdir cfs_unlink

static int cfs_rename(struct inode * old_dir, struct dentry *old_dentry, 
		      struct inode * new_dir, struct dentry *new_dentry)
{
	int retval;
	
	retval = ll_cfs_rename(old_dir, old_dentry, new_dir, new_dentry);

	if (!retval) {
		struct inode *inode = new_dentry->d_inode;
		if (inode) {
			inode->i_nlink--;
			dput(new_dentry);
		}
	}
	return retval;
}

static int cfs_dir_revalidate(struct dentry *dentry)
{
        printk("CFS: cfs_dir_revalidate: %s\n", dentry->d_name.name);
        return (!cfs_lookup_revalidate(dentry, 0));
}


static int cfs_file_revalidate(struct dentry *dentry)
{
	printk("CFS: cfs_file_revalidate: %s\n", dentry->d_name.name);
	truncate_inode_pages(dentry->d_inode->i_mapping, 0);
	return 0;
}

static int cfs_symlink(struct inode * dir, struct dentry *dentry, const char * symname)
{
	int error;
	
	error =  ll_cfs_symlink(dir, dentry, symname);
	
	return error;
}

static int cfs_symlink_filler(struct file *file, struct page *page)
{
	struct inode *inode = page->mapping->host;
	struct dentry *dentry = list_entry(inode->i_dentry.next, struct dentry, d_alias);
        int error;
        unsigned int len = PAGE_SIZE;
        char *p = kmap(page);

        error = ll_cfs_readlink(dentry, p, &len);
	
        if (error)
                goto fail;
        SetPageUptodate(page);
        kunmap(page);
        UnlockPage(page);
        return 0;

fail:
        SetPageError(page);
        kunmap(page);
        UnlockPage(page);
        return error;
}

static int cfs_readdir(struct file * filp, void * dirent, filldir_t filldir)
{
	int error;

	error= ll_cfs_readdir(filp, dirent, filldir);
	
	return error;
}


static int cfs_setattr(struct dentry * dentry, struct iattr * attr)
{
	int error = 0;
	
	if (!(error = ll_cfs_setattr(dentry, attr))) {
		
		return inode_setattr(dentry->d_inode, attr);
	}
	
	return error;
}

static int cfs_sync_file(struct file * file, struct dentry *dentry, int datasync)
{
        return 0;
}

/***********************************************************/

static struct address_space_operations cfs_aops = {
	readpage:	cfs_readpage,
	writepage:      fail_writepage,
/*	writepage:	cfs_writepage, */
	prepare_write:  cfs_prepare_write,
	commit_write:   cfs_commit_write
};

struct address_space_operations cfs_symlink_aops = {
        readpage:       cfs_symlink_filler
};

struct inode_operations cfs_symlink_inode_operations = {
        readlink:       page_readlink,
        follow_link:    page_follow_link,
/*        revalidate:     cfs_file_revalidate, */
        setattr:        cfs_setattr,
};

static struct file_operations cfs_file_operations = {
	read:		generic_file_read,
	write:		generic_file_write,
/*	write:          cfs_file_write, */
	mmap:		generic_file_mmap,
	fsync:		cfs_sync_file,
};

static struct inode_operations cfs_file_inode_operations = {
	create:		cfs_create,
	revalidate :    cfs_file_revalidate,
	setattr:        cfs_setattr,
};

static struct file_operations cfs_dir_operations = {
	read:		generic_read_dir,
	readdir:	cfs_readdir, /* dcache_readdir, */
	fsync:		cfs_sync_file,
};

static struct inode_operations cfs_dir_inode_operations = {
	create:		cfs_create,
	lookup:		cfs_lookup,
	link:		cfs_link,
	unlink:		cfs_unlink,
	symlink:	cfs_symlink,
	mkdir:		cfs_mkdir,
	rmdir:		cfs_rmdir,
	mknod:		cfs_mknod,
	rename:		cfs_rename,
	setattr:        cfs_setattr,
/*	revalidate:     cfs_dir_revalidate, */
};

struct dentry_operations cfs_dentry_operations = {
	d_revalidate:   cfs_lookup_revalidate,
};

static struct super_operations cfs_ops = {
/*	read_inode:     cfs_read_inode, */
	put_inode:	force_delete,
	statfs:		cfs_statfs,
};

struct super_block *cfs_read_super(struct super_block * sb, void * data, int silent)
{
	struct inode * inode;
	struct dentry * root;

	if (ll_cfs_read_super(sb, data, silent))
		return NULL;

	sb->s_blocksize = PAGE_CACHE_SIZE;
	sb->s_blocksize_bits = PAGE_CACHE_SHIFT;
	sb->s_magic = CFS_MAGIC;
	sb->s_op = &cfs_ops;
	inode = cfs_get_inode(sb, S_IFDIR | 0755, 0);
	if (!inode)
		return NULL;

	root = d_alloc_root(inode);
	if (!root) {
		iput(inode);
		return NULL;
	}
	sb->s_root = root;
	sb->s_root->d_op = &cfs_dentry_operations;
	
	CFS_ALLOC(sb->s_root->d_fsdata, strlen(data)+1);
	mempcy(sb->s_root->d_fsdata, data, strlen(data)+1);

	return sb;
}

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
