/*
 * CFS (Corba File System) for Linux.
 * Livio Soares <livio@ime.usp.br>
 * http://www.lcpd.ime.usp.br/~livio/CFS/
 *   
 * This was adapted from CODA's psdev.c to become CFS's psdev
 */

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/lp.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/poll.h>
#include <asm/uaccess.h>

#include <linux/cfs.h>

struct upcall_request {
	struct list_head     uc_chain;
	struct cfs_upcall   *uc_up;
	struct cfs_downcall *uc_down;
	wait_queue_head_t    uc_sleep;
	unsigned long        uc_unique;
};

struct comm_queues {
	unsigned long       cq_unique;
	struct list_head    cq_pending;
	struct list_head    cq_processing;
	wait_queue_head_t   cq_waitq;
} cfs_queues;

/* 
 * CFS stuff
 */
extern struct file_system_type cfs_fs_type;

/*
 * Device operations
 */

static unsigned int cfs_psdev_poll(struct file *file, poll_table * wait)
{
        unsigned int mask = POLLOUT | POLLWRNORM;

	poll_wait(file, &cfs_queues.cq_waitq, wait);
        if (!list_empty(&cfs_queues.cq_pending))
                mask |= POLLIN | POLLRDNORM;

        return mask;
}

static int cfs_psdev_ioctl(struct inode * inode, struct file * filp, 
			    unsigned int cmd, unsigned long arg)
{
       unsigned int data;

        switch(cmd) {
        case CIOC_KERNEL_VERSION:
                data = CFS_KERNEL_VERSION;
                return put_user(data, (int *) arg);
        default:
                return -ENOTTY;
        }
	return 0;
}

/*
 *	Receive a message written by CFSd to the psdev
 */
 
static ssize_t cfs_psdev_write(struct file *file, const char *buf, 
				size_t nbytes, loff_t *off)
{
	struct upcall_request *req = NULL, *tmp;
	struct cfs_downcall *down;
	struct list_head *lh;

	CFS_ALLOC(down, struct cfs_downcall *, sizeof(struct cfs_downcall)); 
	if (copy_from_user(down, buf, sizeof(struct cfs_downcall)))
		return -EFAULT;

	/* Look for the message on the processing queue. */
        list_for_each(lh, &cfs_queues.cq_processing) {
                tmp = list_entry(lh, struct upcall_request, uc_chain);
                if (tmp->uc_unique == down->unique) {
                        req = tmp;
                        list_del(&req->uc_chain);
                        break;
                }
        }

	/* Found it! Copy necessary info to the downcall, wake up the caller
	   so that he can resume his work in ll_cfs_*  */
	if (req) {
		memcpy(req->uc_down, down, sizeof(struct cfs_downcall));
		CFS_FREE(down, sizeof(struct cfs_downcall));

		/* process here all the 2-fase memory transfers*/
		switch (req->uc_down->type) {
		case CFS_READDIR: 
		{
			CFS_ALLOC(req->uc_down->u.readdir.dirp, struct dirent *,
				  req->uc_down->u.readdir.ndir);
			if (copy_from_user(req->uc_down->u.readdir.dirp, buf+sizeof(struct cfs_downcall), 
					   req->uc_down->u.readdir.ndir))
				return -EFAULT;
			break;
		}
		case CFS_READPAGE:
		{
			CFS_ALLOC(req->uc_down->u.readpage.buf, char *,
				  req->uc_down->u.readpage.nbytes);
			if (copy_from_user(req->uc_down->u.readpage.buf, buf+sizeof(struct cfs_downcall),
					   req->uc_down->u.readpage.nbytes))
				return -EFAULT;
			break;
		}
		case CFS_READLINK:
		{
			CFS_ALLOC(req->uc_down->u.readlink.buf, char *,
				  req->uc_down->u.readlink.nbytes);
			if (copy_from_user(req->uc_down->u.readlink.buf, buf+sizeof(struct cfs_downcall),
					   req->uc_down->u.readlink.nbytes))
				return -EFAULT;
			break;
		}
		default:
			break;
		}

		wake_up(&req->uc_sleep);
		return nbytes;
	}
	CFS_FREE(down, sizeof(struct cfs_downcall));
	printk("CFS: psdev_write: PAU!!\n");
	return 0;
}

/*
 *	Read a message from the kernel to Venus
 */

static ssize_t cfs_psdev_read(struct file * file, char * buf, 
			       size_t nbytes, loff_t *off)
{
	DECLARE_WAITQUEUE(wait, current);
	struct upcall_request *req;
	ssize_t retval = 0;

	add_wait_queue(&cfs_queues.cq_waitq, &wait);
        set_current_state(TASK_INTERRUPTIBLE);
	
	while (list_empty(&cfs_queues.cq_pending)) {
		if (file->f_flags & O_NONBLOCK) {
			retval = -EAGAIN;
			break;
		}
		if (signal_pending(current)) {
                        retval = -ERESTARTSYS;
                        break;
                }
                schedule();
        }

        set_current_state(TASK_RUNNING);
        remove_wait_queue(&cfs_queues.cq_waitq, &wait);

	if (retval < 0)
		goto out;

	req = list_entry(cfs_queues.cq_pending.next, struct upcall_request, uc_chain);
	list_del(&(req->uc_chain));

	copy_to_user(buf, (req->uc_up), sizeof(struct cfs_upcall));
	retval = sizeof(struct cfs_upcall);

	/* process here all the 2-fase memory transfers*/
	switch (req->uc_up->type){
	case CFS_COMMIT_WRITE: 
		retval += req->uc_up->u.commitwrite.nbytes;
		if (copy_to_user(buf+sizeof(struct cfs_upcall), req->uc_up->u.commitwrite.buf, 
				   req->uc_up->u.commitwrite.nbytes))
			return -EFAULT;
		break;
	default:
		break;
	}

	list_add(&(req->uc_chain), cfs_queues.cq_processing.prev);

   out:
	return retval;
}

static int cfs_psdev_open(struct inode * inode, struct file * file)
{
	return 0;
}


static int cfs_psdev_release(struct inode * inode, struct file * file)
{
	return 0;
}


static struct file_operations cfs_psdev_fops = {
	owner:		THIS_MODULE,
	read:		cfs_psdev_read,
	write:		cfs_psdev_write,
	poll:		cfs_psdev_poll,
	ioctl:		cfs_psdev_ioctl,
	open:		cfs_psdev_open,
	release:	cfs_psdev_release,
};

/****************************************************************/

int psdev_enqueue(struct cfs_upcall *up, struct cfs_downcall *down)
{
	struct upcall_request req;
	DECLARE_WAITQUEUE(wait, current);

	req.uc_up = up;
	req.uc_down = down;
	req.uc_down->unique = req.uc_up->unique = req.uc_unique = cfs_queues.cq_unique++;
	init_waitqueue_head(&req.uc_sleep);

	list_add(&(req.uc_chain), cfs_queues.cq_pending.prev);

	/* wake up processes listening on cfs/X for requests (CFSd) */
        wake_up_interruptible(&cfs_queues.cq_waitq);

	/* now I want to wait here until a downcall comes in 
         * that matches the sequence number.  i stay on the waitqueue until 
         * I explicitly remove myself.
         */
        add_wait_queue(&req.uc_sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);
	schedule();
	remove_wait_queue(&req.uc_sleep, &wait);
	set_current_state(TASK_RUNNING);

	return 0;
}

/****************************************************************/


extern devfs_handle_t devfs_handle;

int init_cfs_psdev(void)
{
	if(devfs_register_chrdev(CFS_PSDEV_MAJOR,"cfs_psdev",
				 &cfs_psdev_fops)) {
              printk(KERN_ERR "cfs_psdev: unable to get major %d\n", 
		     CFS_PSDEV_MAJOR);
              return -EIO;
	}
	devfs_handle = devfs_mk_dir (NULL, "cfs", NULL);
	devfs_register_series (devfs_handle, "%u", MAX_CFSDEVS, DEVFS_FL_NONE,
			       CFS_PSDEV_MAJOR, 0,
			       S_IFCHR | S_IRUSR | S_IWUSR,
			       &cfs_psdev_fops, NULL);

	INIT_LIST_HEAD(&cfs_queues.cq_pending);
	INIT_LIST_HEAD(&cfs_queues.cq_processing);
	init_waitqueue_head(&cfs_queues.cq_waitq);
	cfs_queues.cq_unique = 0;

	return 0;
}


static DECLARE_FSTYPE(cfs_fs_type, "cfs", cfs_read_super, FS_LITTER);

static devfs_handle_t devfs_handle;

static int __init init_cfs_fs(void)
{
	int status;
	printk(KERN_INFO "CFS, v0.0.1\n");

	status = init_cfs_psdev();
	if ( status ) {
		printk("Problem (%d) in init_cfs_psdev\n", status);
		return status;
	}
	
	status = register_filesystem(&cfs_fs_type);
	if (status) {
		printk("cfs: failed to register filesystem!\n");
		devfs_unregister(devfs_handle);
		devfs_unregister_chrdev(CFS_PSDEV_MAJOR,"cfs_psdev");
	}
	return status;
}

static void __exit exit_cfs_fs(void)
{
	int err;

	err = unregister_filesystem(&cfs_fs_type);
        if ( err != 0 ) {
                printk("cfs: failed to unregister filesystem!\n");
        }
	devfs_unregister(devfs_handle);
	devfs_unregister_chrdev(CFS_PSDEV_MAJOR, "cfs_psdev");
}

MODULE_AUTHOR("Livio Baldini Soares <livio@ime.usp.br> Mrcio Rodrigo de Freitas Carneiro <carneiro@ime.usp.br> Roberto Pires de Carvalho <carvalho@ime.usp.br>");
MODULE_LICENSE("GPL");

module_init(init_cfs_fs)
module_exit(exit_cfs_fs)


/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
