/*
 * CFS (Corba File System) for Linux.
 * Livio Soares <livio@ime.usp.br>
 * http://www.lcpd.ime.usp.br/~livio/CFS/
 *
 * This file is released under the GPL.
 */

#ifndef __CFS_H
#define __CFS_H

#include <linux/vmalloc.h>      /* for vmalloc() */
#include <linux/slab.h>         /* for kmalloc() */
#include <linux/dirent.h>       /* for struct dirent */

/* some random number */
#define CFS_KERNEL_VERSION   1
#define CFS_MAGIC	     0x69696969
#define CFS_PSDEV_MAJOR      69
#define MAX_CFSDEVS          5    /* how many do we allow */
#define CFS_NAME_LEN         255

int init_cfs_psdev(void);
void cfs_read_inode(struct inode *inode);
struct super_block *cfs_read_super(struct super_block * sb, void * data, int silent);
struct inode *cfs_get_inode(struct super_block *sb, int mode, int dev);
void cfs_get_inode_define_mode(struct inode *inode);

struct cfs_downcall {
	unsigned int type;    /* might be unneeded? */
	unsigned int unique;
	int error;

	union {
		struct statfs       statfs;  /* stafs       */
		struct stat           stat;  /* read_inode  */
		struct {
			unsigned int    ndir;
			struct dirent   *dirp;
		} readdir;  /* readdir */
		struct {
			unsigned int    nbytes;
			char              *buf;
		} readpage;
		/* commitwrite null */
		/* mknod       null */
		/* unlink      null */
		/* rename      null */
		struct {
			int              nbytes;
			char             *buf;
		} readlink;
		/* setattr     null */
	} u;
};

struct cfs_upcall {
	unsigned int type;
	unsigned int unique;

	union {
		struct {
			/* null */
		} statfs; 
		struct {
			char inodename[CFS_NAME_LEN];  /* read_inode */
		} readinode;
		struct {
			char dirname[CFS_NAME_LEN];
			unsigned long  file_offset;
		} readdir;
		struct {
			char filename[CFS_NAME_LEN];
			unsigned long file_offset;     /* readpage   */
		} readpage;
		struct {
			char             filename[CFS_NAME_LEN];
			unsigned long    file_offset;
			unsigned int     nbytes;
			char             *buf;
		} commitwrite;                        /* commitwrite */
		struct {
			char             inodename[CFS_NAME_LEN];
			int              mode;
			int              dev;
		} mknod;	      	              /* mknod  */
		struct {
			char             inodename[CFS_NAME_LEN];
			int              mode;
		} unlink;                            /* unlink  */
		struct {
			char             oldpath[CFS_NAME_LEN];
			char             newpath[CFS_NAME_LEN];
		} rename;                           /*  rename  */
		struct {
			char             linkname[CFS_NAME_LEN];
		} readlink;                         /* readlink */
		struct {
			char             oldpath[CFS_NAME_LEN];
			char             newpath[CFS_NAME_LEN];
		} symlink;                          /* symlink  */
		struct {
			char             inodename[CFS_NAME_LEN];
			struct iattr     attr;
		} setattr;                          /* setattr  */

	} u;

};

#define upcall_alloc() kmalloc(sizeof(struct cfs_upcall), GFP_KERNEL)
#define upcall_free(r) kfree(r)

/* Low-Level upcall stubs */
int ll_cfs_statfs(struct super_block *sb, struct statfs *buf);
/*int ll_cfs_read_inode(struct inode *inode); */
int ll_cfs_lookup(struct inode **inode, struct dentry *dentry);
int ll_cfs_readpage(struct dentry *dentry, loff_t offset, char * buffer);
ssize_t ll_cfs_file_write(struct file *file,const char *buf,size_t count,loff_t pos);
int ll_cfs_writepage(struct page *page);
int ll_cfs_prepare_write(struct file *file, struct page *page, unsigned offset, unsigned to);
int ll_cfs_commit_write(struct file *file, struct page *page, unsigned offset, unsigned to);
void ll_cfs_force_delete(struct inode *inode);
int ll_cfs_mknod(struct inode *dir, struct dentry *dentry, int mode, int dev);
int ll_cfs_mkdir(struct inode * dir, struct dentry * dentry, int mode);
int ll_cfs_create(struct inode *dir, struct dentry *dentry, int mode);
int ll_cfs_link(struct dentry *old_dentry, struct inode * dir, struct dentry * dentry);
inline int ll_cfs_positive(struct dentry *dentry);
int ll_cfs_empty(struct dentry *dentry);
int ll_cfs_unlink(struct inode * dir, struct dentry *dentry);
int ll_cfs_rmdir(struct inode * dir, struct dentry *dentry);
int ll_cfs_rename(struct inode * old_dir, struct dentry *old_dentry, struct inode * new_dir,struct dentry *new_dentry);
int ll_cfs_symlink(struct inode * dir, struct dentry *dentry, const char * symname);
int ll_cfs_readlink(struct dentry *dentry, char *page, int *len);
int ll_cfs_sync_file(struct file * file, struct dentry *dentry, int datasync);
int ll_cfs_readdir(struct file * filp, void * dirent, filldir_t filldir);
int ll_cfs_setattr(struct dentry *dentry, struct iattr * attr);
struct super_block *ll_cfs_read_super(struct super_block * sb, void * data, int silent);

/* Kernel <-> CORBA daemon communication */
#define CFS_STATFS            2
#define CFS_LOOKUP            3
#define CFS_READPAGE          4
#define CFS_FILE_WRITE        5
#define CFS_WRITEPAGE         6
#define CFS_PREPARE_WRITE     7
#define CFS_COMMIT_WRITE      8
#define CFS_READ_INODE        9
#define CFS_FORCE_DELETE     10
#define CFS_MKNOD            11
#define CFS_MKDIR            12
#define CFS_CREATE           13
#define CFS_LINK             14
#define CFS_POSITIVE         15
#define CFS_EMPTY            16
#define CFS_UNLINK           17
#define CFS_RMDIR            18
#define CFS_RENAME           19
#define CFS_SYMLINK          20
#define CFS_READLINK         21
#define CFS_SYNC_FILE        22
#define CFS_READDIR          23
#define CFS_SETATTR          24
#define CFS_READ_SUPER       25

int psdev_enqueue(struct cfs_upcall *up, struct cfs_downcall *down);


#define CFS_ALLOC(ptr, cast, size)                                       \
do {                                                                      \
    if (size < PAGE_SIZE)                                                 \
        ptr = (cast)kmalloc((unsigned long) size, GFP_KERNEL);            \
     else                                                                 \
        ptr = (cast)vmalloc((unsigned long) size);                        \
    if (ptr == 0)                                                         \
        printk("kernel malloc returns 0 at %s:%d\n", __FILE__, __LINE__); \
    else memset( ptr, 0, size );                                          \
} while (0)


#define CFS_FREE(ptr,size)                                               \
do {                                                                      \
	if (size < PAGE_SIZE)                                             \
		kfree((ptr));                                             \
	else                                                              \
		vfree((ptr));                                             \
} while (0)


#endif /* __CFS_PSDEV_H */

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
